from dragonk8s.lib.client.tools.record import EventSink
from kubernetes.client.models.core_v1_event import CoreV1Event
from kubernetes import client


class EventSinkImpl(EventSink):

    def __init__(self):
        super().__init__()
        self.client = client.CoreV1Api()

    def create(self, event: CoreV1Event) -> CoreV1Event:

        return self.client.create_namespaced_event(event.metadata.namespace, event)

    def update(self, event: CoreV1Event) -> CoreV1Event:
        return self.client.replace_namespaced_event(event.metadata.name, event.metadata.namespace, event)

    def patch(self, old_event: CoreV1Event, data) -> CoreV1Event:
        return self.client.patch_namespaced_event(old_event.metadata.name, old_event.metadata.namespace, data)



import datetime
import socket
import time

import copy
import json
import logging
from threading import Event, Lock, Thread
from kubernetes.client.api.events_v1_api import EventsV1Api
from kubernetes.client.models.events_v1_event import EventsV1Event
from dragonk8s.lib.apimachinery.pkg.watch import watch, mux
from kubernetes.client.exceptions import ApiException
from kubernetes.client.models.events_v1_event_series import EventsV1EventSeries
from dragonk8s.lib.apimachinery.pkg.util import wait
from dragonk8s.lib.client.tools import reference
from kubernetes.client.models.v1_object_reference import V1ObjectReference
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from dragonk8s.lib.util import timeutil
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.apimachinery.pkg.types import patch
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager


finishTime = 1 * 60
refreshTime = 30

EventTypeNormal = "Normal"
EventTypeWarning = "Warning"


def get_event_string(event: EventsV1Event, indent=4) -> str:
    res = {}
    for k, v in event.to_dict().items():
        if k.startswith("deprecated_"):
            continue
        if k == "managedFields":
            continue
        res[k] = v
    return json.dumps(res, indent=indent)


def validate_event_type(eventtype: str) -> bool:
    if eventtype == EventTypeNormal or eventtype == EventTypeWarning:
        return True
    return False


def is_key_not_found_error(ex) -> bool:
    return ex is not None and ex.status == 404


class EventRecorder(object):

    def eventf(self, regarding, related, eventtype: str, reason: str, action: str, note: str, *args):
        pass


class EventBroadcaster(object):

    def start_recording_to_sink(self, stop_event: Event):
        pass

    def new_recorder(self, reporting_controller: str) -> EventRecorder:
        pass

    def start_event_watcher(self, event_handler, stop_event: Event):
        pass

    def start_structured_logging(self, stop_event: Event):
        pass

    def shut_down(self):
        pass


class EventSink(object):

    def create(self, event: EventsV1Event) -> EventsV1Event:
        return EventsV1Event()

    def update(self, event: EventsV1Event) -> EventsV1Event:
        return EventsV1Event()

    def patch(self, event: EventsV1Event, data) -> EventsV1Event:
        return EventsV1Event()


def ref_2_str(ref):
    if ref is None:
        return ""
    if isinstance(ref, V1ObjectReference):
        ref = ref.to_dict()
    return json.dumps(ref)


class EventKey(object):

    def __init__(self, action: str = "", reason: str = "", reporting_controller: str = "", regarding=None,
                 related=None):
        self.action = action
        self.reason = reason,
        self.reporting_controller = reporting_controller
        self.regarding = regarding
        self.related = related

    def __str__(self):
        return "action={},reason={},reporting_controller={},regarding={},related={}".format(
            self.action, self.reason, self.reporting_controller, ref_2_str(self.regarding), ref_2_str(self.related))

    def __repr__(self):
        return self.__str__()


def _get_key(event: EventsV1Event) -> EventKey:
    key = EventKey(
        action=event.action,
        reason=event.reason,
        reporting_controller=event.reporting_controller,
        regarding=event.regarding,
    )
    if event.related is not None:
        key.related = event.related
    return str(key)


class EventSinkImpl(EventSink):

    def __init__(self, kube_client):
        super().__init__()
        self.client = apigvk.get_resource_client(kube_client, apigvk.EventsV1Event)

    def create(self, event: EventsV1Event) -> EventsV1Event:
        if event.metadata is None or event.metadata.namespace == "":
            raise Exception("can't create an event with empty namespace")
        return self.client.create(event)

    def update(self, event: EventsV1Event) -> EventsV1Event:
        if event.metadata is None or event.metadata.namespace == "":
            raise Exception("can't update an event with empty namespace")
        return self.client.replace(body=event, name=event.metadata.name, namespace=event.metadata.namespace)

    def patch(self, event: EventsV1Event, data) -> EventsV1Event:
        if event.metadata is None or event.metadata.namespace == "":
            raise Exception("can't patch an event with empty namespace")
        return self.client.patch(body=data, name=event.metadata.name,
                                                  namespace=event.metadata.namespace,
                                                  content_type=patch.PatchType.StrategicMergePatchType)


def _create_patch_for_series(event: EventsV1Event) -> dict:
    if event.series is None:
        return dict()
    return dict(
        series=dict(
            count=event.series.count,
            lastObservedTime=timeutil.to_time_str_with_ns(event.series.last_observed_time),
        )
    )


def _record_event(sink: EventSink, event: EventsV1Event) -> (EventsV1Event, bool):
    new_event = None
    ex = None
    is_event_series = event.series is not None
    if is_event_series:
        patch = _create_patch_for_series(event)
        try:
            new_event = sink.patch(event, patch)
        except ApiException as e:
            ex = e
    if not is_event_series or (is_event_series and is_key_not_found_error(ex)):
        ex = None
        event.metadata.resource_version = ""
        try:
            new_event = sink.create(event)
        except ApiException as e:
            ex = e
    if ex is None:
        return new_event, False
    logging.error("Unable to write event: %s, err: %s (may retry after sleeping)" % (event.metadata.name, str(ex)))
    return None, True


class RecorderImpl(EventRecorder):

    def __init__(self, reporting_controller: str = "", reporting_instance: str = "", broadcaster:mux.Broadcaster = None):
        self.reporting_controller = reporting_controller
        self.reporting_instance = reporting_instance
        self.broadcaster = broadcaster

    def _make_event(self, ref_regarding: V1ObjectReference, refRelated: V1ObjectReference, timestamp: float, eventtype,
                    reason, message, reporting_controller, reporting_instance, action) -> EventsV1Event:
        namespace = ref_regarding.namespace
        if namespace is None or namespace == "":
            namespace = meta_v1.NamespaceDefault
        return EventsV1Event(
            api_version="events.k8s.io/v1",
            kind="Event",
            metadata=V1ObjectMeta(
                name="%s.%f" % (ref_regarding.name, timestamp),
                namespace=namespace,
            ),
            event_time=timeutil.to_time_str_with_ns(timestamp),
            series=None,
            reporting_controller=reporting_controller,
            reporting_instance=reporting_instance,
            action=action,
            reason=reason,
            regarding=ref_regarding,
            related=refRelated,
            note=message,
            type=eventtype,
        )

    def eventf(self, regarding, related, eventtype: str, reason: str, action: str, note: str, *args):
        timestamp = time.time()
        message = note.format(*args)
        try:
            ref_regarding = reference.get_reference(regarding)
        except Exception as e:
            logging.error("Could not construct reference to %s due to %s, will not report event: '%s' '%s' '%s'"
                          % (str(ref_regarding), str(e), eventtype, reason, message))
            return
        ref_related = None
        if related is not None:
            try:
                ref_related = reference.get_reference(related)
            except Exception as e:
                logging.error("Could not construct reference to %s due to %s, will not report event: '%s' '%s' '%s'"
                              % (str(related), str(e), eventtype, reason, message))
        if not validate_event_type(eventtype):
            logging.error("Unsupported event type: %s" % eventtype)
            return

        event = self._make_event(ref_regarding, ref_related, timestamp, eventtype, reason, message,
                                 self.reporting_controller, self.reporting_instance, action)
        # todo: 异步
        self.broadcaster.action(watch.EventType.Added, event)


class EventBroadcasterImpl(EventBroadcaster):

    def __init__(self, sink: EventSink):
        self.broadcaster = mux.Broadcaster(1000, mux.FullQueueBehavior.DropIfChannelFull)
        self.lock = Lock()
        self.event_cache = dict()
        self.sleep_duration = 10
        self.sink = sink

    def start_recording_event(self, stop_event: Event):
        def handle(obj):
            self.record_to_sink(event=obj)
        stop_watcher = self.start_event_watcher(handle, stop_event)

        # def wait_stop():
        #     while stop_event.wait(10):
        #         pass
        #     print("stop watcher...")
        #     stop_watcher()
        # backend = Thread(target=wait_stop, name="wait_stop")
        # backend.start()

    def start_recording_to_sink(self, stop_event: Event):
        wait.until_with_thread(self._refresh_existing_event_series, refreshTime, stop_event, "_refresh_existing_event_series")
        wait.until_with_thread(self._finish_series, finishTime, stop_event, "_finish_series")
        self.start_recording_event(stop_event)

    def new_recorder(self, reporting_controller: str) -> EventRecorder:
        hostname = socket.gethostname()
        reporting_instance = "{}-{}".format(reporting_controller, hostname)
        return RecorderImpl(
            reporting_controller=reporting_controller,
            reporting_instance=reporting_instance,
            broadcaster=self.broadcaster,
        )

    def start_event_watcher(self, event_handler, stop_event: Event):
        watcher = self.broadcaster.watch()

        def do():
            watch_event = None
            while not stop_event.is_set():
                try:
                    watch_event = None
                    watch_event = watcher.get_result_queue().get(True, 5)
                except:
                    time.sleep(1)
                if watch_event is None:
                    continue
                event_handler(watch_event.obj)
        backend = GlobalThreadManager.new_thread(target=do, generate_name="start_event_watcher")
        backend.start()
        return watcher.stop

    def start_structured_logging(self, stop_event: Event):
        def do(obj: EventsV1Event):
            logging.info(get_event_string(obj))
        self.start_event_watcher(do, stop_event)

    def shut_down(self):
        self.broadcaster.shutdown()

    def _refresh_existing_event_series(self):
        try:
            self.lock.acquire(True)
            for k, event in self.event_cache.items():
                if event.series is not None:
                    recorded_event, retry = _record_event(self.sink, event)
                    if not retry and recorded_event is not None:
                        self.event_cache[k] = recorded_event
        finally:
            self.lock.release()

    def _finish_series(self):
        now = time.time()
        with self.lock:
            to_delete = []
            for k, event in self.event_cache.items():
                event_serie = event.series
                if event_serie is not None:
                    if event_serie.last_observed_time.timestamp() < now - finishTime:
                        recorded_event, retry = _record_event(self.sink, event)
                        if not retry:
                            to_delete.append(k)
                else:
                    if isinstance(event.event_time, str):
                        logging.error("event time is str: %s" % event.event_time)
                        to_delete.append(k)
                    else:
                        if event.event_time.timestamp() < now - finishTime:
                            to_delete.append(k)
            for k in to_delete:
                del self.event_cache[k]

    def _attempt_recording(self, event: EventsV1Event) -> EventsV1Event:
        tries = 0
        while True:
            recorded_event, retry = _record_event(self.sink, event)
            if not retry:
                return recorded_event
            tries += 1
            if tries >= 12:
                logging.error("unable to write event: %s (retry limit exceeded)" % event.metadata.name)
                return None
            time.sleep(self.sleep_duration)

    # todo: 并行
    def record_to_sink(self, event: EventsV1Event):
        event_copy = copy.deepcopy(event)

        def do() -> EventsV1Event:
            try:
                self.lock.acquire(True)
                event_key = _get_key(event_copy)
                if event_key in self.event_cache:
                    event_in_cache = self.event_cache[event_key]
                    if event_in_cache.series is not None:
                        event_in_cache.series.count += 1
                        event_in_cache.series.last_observed_time = datetime.datetime.now()
                        return None
                    event_in_cache.series = EventsV1EventSeries(
                        count=2,
                        last_observed_time=datetime.datetime.now(),
                    )
                    return event_in_cache
                self.event_cache[event_key] = event_copy
                return event_copy
            finally:
                self.lock.release()
        ev_to_record = do()
        if ev_to_record is not None:
            recorded_event = self._attempt_recording(ev_to_record)
            if recorded_event is not None:
                recorded_event_key = _get_key(recorded_event)
                try:
                    self.lock.acquire(True)
                    self.event_cache[recorded_event_key] = recorded_event
                finally:
                    self.lock.release()



import time
import unittest
import logging
from threading import Event
from kubernetes import config
from dragonk8s.lib.client.tools import events
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from dragonk8s.lib.client.kubernetes.typed.core.v1.event_expansion import EventSinkImpl
from dragonk8s.dragon.configuration import Configuration
from dragonk8s.lib.client import CommonClient
from dragonk8s.dragon.api_client import ApiClient

class EventsTest(unittest.TestCase):

    def setUp(self):
        cfg = Configuration()
        config.load_kube_config(config_file="C:\\Users\\Administrator\\.kube\\config", client_configuration=cfg)
        Configuration.set_default(cfg)
        self.base_client = ApiClient()
        client = CommonClient(self.base_client)

        sink = events.EventSinkImpl(client)
        self.event_broadcaster = events.EventBroadcasterImpl(sink=sink)
        self.event_recorder = self.event_broadcaster.new_recorder("mhc-controller")
        self.stop = Event()
        print(self.stop.is_set())
        self.event_broadcaster.start_recording_to_sink(self.stop)
        self.event_broadcaster.start_structured_logging(self.stop)

    def test1(self):
        pod = V1Pod(
            kind="Pod",
            api_version="v1",
            metadata=V1ObjectMeta(
                name="bbb",
                namespace="default"
            ),
        )
        self.event_recorder.eventf(pod, None, events.EventTypeNormal, "MHCReason", "mhc-action", "some message....")
        time.sleep(5)

    def test2(self):
        pod = V1Pod(
            kind="Pod",
            api_version="v1",
            metadata=V1ObjectMeta(
                name="bbb",
                namespace="default"
            ),
        )
        for i in range(5):
            self.event_recorder.eventf(pod, None, events.EventTypeNormal, "MHCReason", "mhc-action", "some message....")
            time.sleep(1)
        time.sleep(90)

    def tearDown(self) -> None:
        print("shut down")
        self.stop.set()
        self.event_broadcaster.shut_down()



import copy

import json
import re
from kubernetes.client.models.v1_status import V1Status
from kubernetes.client.models.v1_status_details import V1StatusDetails
from kubernetes.client.models.v1_list_meta import V1ListMeta
from kubernetes.client.models.v1_status_cause import V1StatusCause
from kubernetes.client.api_client import ApiClient

Attribute2Obj = {
    "V1Status": V1Status(),
    "V1StatusDetails": V1StatusDetails(),
    "V1ListMeta": V1ListMeta(),
    "V1StatusCause": V1StatusCause(),
}

client = ApiClient()

class Resp(object):

    def __init__(self, data):
        self.data = data

def turn_param_style(params: dict):
    """
    将参数名的驼峰形式转为下划线形式
    @param params:
    @return:
    """
    temp_dict = {}
    for name, value in params.items():
        temp_name = ""
        if re.search("[A-Z]", name):
            capital_letters = re.findall("[A-Z]", name)
            for c in capital_letters:
                lower_c = c.lower()
                r_str = "_" + lower_c
                temp_name = name.replace(c, r_str)
        else:
            temp_name = name

        temp_dict.update({temp_name: value})

    return temp_dict


def _json_to_py_object(data, obj):
    """
    将json格式的数据转成python中k8s的对象
    :param data:
    :param obj:
    :return:
    """
    obj_copy = obj
    if data is None:
        return obj_copy
    if isinstance(data, dict):
        data_dict = data
    elif isinstance(data, str):
        data_dict = json.loads(data)
    else:
        return obj_copy
    att_map_reverse = {v: k for k, v in getattr(obj_copy, "attribute_map").items()}
    for k, v in data_dict.items():
        if k not in att_map_reverse:
            continue
        py_k = att_map_reverse[k]
        my_type = getattr(obj_copy, "openapi_types")[py_k]
        patt = "list\[(.*)\]"
        is_list = False
        m = re.match(patt, my_type)
        if m is not None and len(m.group())> 1:
            is_list = True
            my_type = m.groups([0])
        if my_type in {"str", "int", "datetime"}:
            setattr(obj_copy, py_k, v)
            continue
        if is_list:
            setattr(obj_copy, py_k, json_to_py_object_list(v, get_simple_obj_by_name(my_type)))
        else:
            setattr(obj_copy, py_k, _json_to_py_object(v, get_simple_obj_by_name(my_type)))

    return obj_copy


def json_to_py_object_list(data, obj):
    """
    将json格式的数据转成python中k8s的对象
    :param data:
    :param obj:
    :return:
    """
    if data is None:
        return []
    if isinstance(data, list):
        data_list = data
    else:
        data_list = json.loads(data)

    res = []
    for d in data_list:
        res.append(_json_to_py_object(d, obj))
    return res


def parse_v1_status(data: str) -> V1Status:
    status = V1Status()
    _json_to_py_object(data, status)
    status.details = _json_to_py_object(status.details, V1StatusDetails())
    status.metadata = _json_to_py_object(status.metadata, V1ListMeta())
    status.details.causes = json_to_py_object_list(status.details.causes, V1StatusCause())
    return status


def get_simple_obj_by_name(name: str):
    if name in Attribute2Obj:
        return copy.deepcopy(Attribute2Obj[name])
    raise Exception("%s is not in Attribute2Obj cache" % name)


def parse_json_to_object(data: str, obj):
    if isinstance(obj, str):
        obj = get_simple_obj_by_name(obj)
    return _json_to_py_object(data, obj)


def parse_json_to_object_by_class_name(data: str, obj_type):
    return client.deserialize(Resp(data), obj_type)



import functools
import logging
import queue
import threading
import time
from threading import Event
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.apimachinery.pkg.runtime import schema
from dragonk8s.lib.client.tools import cache
from dragonk8s.lib.apimachinery.pkg.watch import watch
from dragonk8s.lib.client.tools import pager
from dragonk8s.lib.apimachinery.pkg.api import errors
from kubernetes.client.exceptions import ApiException
from dragonk8s.lib.client.tools import fifo
from dragonk8s.lib.apimachinery.pkg.util import wait
from dragonk8s.lib.client.tools import delta
from queue import Queue
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager


class ResourceEventHandler(object):

    def on_add(self, obj):
        pass

    def on_update(self, old_obj, new_obj):
        pass

    def on_delete(self, obj):
        pass


class ResourceEventHandlerFuncs(ResourceEventHandler):

    def __init__(self, add_func=None, update_func=None, delete_func=None):
        self.add_func = add_func
        self.update_func = update_func
        self.delete_func = delete_func

    def on_add(self, obj):
        if self.add_func:
            self.add_func(obj)

    def on_update(self, old_obj, new_obj):
        if self.update_func:
            self.update_func(old_obj, new_obj)

    def on_delete(self, obj):
        if self.delete_func:
            self.delete_func(obj)


class FilteringResourceEventHandler(ResourceEventHandler):

    def __init__(self, filter_func, handler: ResourceEventHandler):
        self.filter_func = filter_func
        self.handler = handler

    def on_add(self, obj):
        if not self.filter_func(obj):
            return
        self.handler.on_add(obj)

    def on_update(self, old_obj, new_obj):
        newer = self.filter_func(new_obj)
        older = self.filter_func(old_obj)
        if newer and older:
            self.handler.on_update(old_obj, new_obj)
        elif newer and not older:
            self.handler.on_add(new_obj)
        elif not newer and older:
            self.handler.on_delete(old_obj)

    def on_delete(self, obj):
        if not self.filter_func(obj):
            return
        self.handler.on_delete(obj)


class Controller(object):

    def run(self, stop: Event):
        pass

    def has_synced(self) -> bool:
        pass

    def last_sync_resource_version(self) -> str:
        pass


class Lister(object):

    def list(self, options: meta_v1.ListOptions):
        pass


class Watcher(object):

    def watch(self, options: meta_v1.ListOptions) -> watch.Interface:
        pass


class ListerWatcher(Lister, Watcher):
    pass


class ResourceVersionUpdater(object):

    def update_resource_version(self, resource_version: str):
        pass


class Reflector(object):

    def __init__(self, name, lw: ListerWatcher, resync_period=0, store: cache.Store = None, parse_func=None):
        self.name = name
        if not self.name:
            self.name = "inner_controller"
        self.expected_type = None
        self.expected_type_name = ""
        self.expected_gvk = None
        self.store = store
        self.lister_watcher = lw
        # todo
        self.backoff_manager = None
        self.init_conn_backoff_manager = None

        self.resync_period = resync_period
        self.should_resync = None
        self.paginated_result = False
        self.last_sync_resource_version = ""
        self.is_last_sync_resource_version_unavailable = False
        self.last_sync_resource_version_mutex = threading.Lock()
        self.watch_list_page_size = 0
        # todo
        self.watch_error_handler = None
        self.parse_func = parse_func

    def get_last_sync_resource_version(self):
        try:
            self.last_sync_resource_version_mutex.acquire(True)
            return self.last_sync_resource_version
        finally:
            self.last_sync_resource_version_mutex.acquire()

    def set_last_sync_resource_version(self, v: str):
        with self.last_sync_resource_version_mutex:
            self.last_sync_resource_version = v

    def set_last_sync_resource_version_unavailable(self, is_unavailable: bool):
        with self.last_sync_resource_version_mutex:
            self.last_sync_resource_version = is_unavailable

    def _set_expected_type(self, obj):
        self.expected_type = type(obj)
        self.expected_type_name = self.expected_type.__name__
        self.expected_gvk = schema.from_obj(obj)

    def _relist_resource_version(self):
        try:
            self.last_sync_resource_version_mutex.acquire(blocking=True)
            if self.is_last_sync_resource_version_unavailable:
                return ""
            if self.last_sync_resource_version == "":
                return "0"
            return self.last_sync_resource_version
        finally:
            self.last_sync_resource_version_mutex.release()

    def watch_handler(self, start: float, w, resource_version: str, stop: Event):
        new_resource_version = resource_version
        if not w:
            return new_resource_version
        event_count = 0
        while True:
            if stop.is_set():
                return
            event = w.get_result_queue().get(True, 5)
            if event is None:
                continue
            if event['type'] == watch.EventType.End:
                break
            if event['type'] == watch.EventType.Error:
                raise Exception("get error event: " + event['object'].to_str())
            if self.parse_func:
                obj = self.parse_func(event['object'])
            else:
                obj = event['object']
            if self.expected_type:
                if not isinstance(obj, self.expected_type):
                    logging.error("%s expected type %s, but watch event object had type %s" % (
                        self.name, self.expected_type.__name__, type(obj).__name__))
                    continue
            if self.expected_gvk:
                gvk = schema.from_obj(obj)
                if str(self.expected_gvk) != str(gvk):
                    logging.error("%s expected gvk %s, but watch event object had gvk %s" % (
                        self.name, str(self.expected_gvk), str(gvk)))
                    continue
            if not hasattr(obj, "metadata"):
                logging.error("%s: unable to understand watch event: %s" % (self.name, str(event)))
                continue
            new_resource_version = obj.metadata.resource_version
            if event['type'] == watch.EventType.Added:
                self.store.add(obj)
            elif event['type'] == watch.EventType.Modified:
                self.store.update(obj)
            elif event['type'] == watch.EventType.Deleted:
                self.store.delete(obj)
            elif event['type'] == watch.EventType.Bookmark:
                pass
            else:
                logging.error("%s: unable to understand watch event: %s" % (self.name, str(event)))
            self.set_last_sync_resource_version(new_resource_version)

            if hasattr(self.store, "update_resource_version"):
                self.store.update_resource_version(new_resource_version)
            event_count += 1
        watch_duration = time.time() - start
        if watch_duration < 60 and event_count == 0:
            raise Exception(
                "very short watch: %s : Unexpected watch close - watch lasted less than a second and no items received"
                % self.name)
        logging.info("%s: Watch close - %s total %d items received" % (self.name, self.expected_type_name, event_count))
        return new_resource_version

    def _sync_with(self, items, resource_version: str):
        self.store.replace(items, resource_version)

    def list_and_watch(self, stop: Event):
        logging.info("listing and watch %s from %s", self.expected_type_name, self.name)
        resource_version = "0"
        options = meta_v1.ListOptions(
            resource_version=self._relist_resource_version()
        )
        ex = None
        try:
            paginated_result = False
            res_list = None

            page = pager.ListPager(pager.simple_page_func(self.lister_watcher.list))
            if self.watch_list_page_size != 0:
                page.page_size = self.watch_list_page_size
            elif self.paginated_result:
                pass
            elif options.resource_version != "" and options.resource_version != "0":
                page.page_size = 0
            try:
                res_list, paginated_result = page.list(options, stop)
            except ApiException as e:
                logging.error("page list error: %s" % str(e))
                if errors.is_resource_expired(e) or errors.is_too_large_resource_version_error(e):
                    self.set_last_sync_resource_version_unavailable(True)
                    rest_list, paginated_result = page.list(
                        meta_v1.ListOptions(resource_version=self._relist_resource_version()), stop)
        except Exception as e:
            logging.error("%s: failed to list %s: %s", self.name, self.expected_type_name, str(e))
            ex = e
        else:
            if options.resource_version == "0" and paginated_result:
                self.paginated_result = True
            self.set_last_sync_resource_version_unavailable(False)
            if res_list is None or not hasattr(res_list, "metadata"):
                logging.error("unable to understand list result: %s" % str(res_list))
            else:
                try:
                    resource_version = res_list.metadata.resource_version
                    items = res_list.items
                    self._sync_with(items, resource_version)
                    self.set_last_sync_resource_version(resource_version)
                except Exception as e:
                    ex = e
        if ex is not None:
            raise ex

        def resync():
            count = 0
            while not stop.is_set():
                time.sleep(1)
                count += 1
                if count >= self.resync_period:
                    count = 0
                    if self.should_resync is None or (self.should_resync is not None and self.should_resync()):
                        logging.info("%s: forcing resync" % self.name)
                        self.store.resync()

        th = GlobalThreadManager.new_thread(target=resync, generate_name="resync")
        th.start()

        while not stop.is_set():
            timeout_seconds = 1
            options = meta_v1.ListOptions(
                resource_version=resource_version,
                timeout_seconds=timeout_seconds,
                allow_watch_bookmarks=True,
            )
            start = time.time()
            try:
                w = self.lister_watcher.watch(options)
            except ApiException as e:
                if errors.is_too_many_requests(e):
                    time.sleep(1)
                    continue
                else:
                    raise e
            except Exception as e:
                if "Connection refused" in str(e):
                    time.sleep(1)
                    continue
                else:
                    raise e
            else:
                try:
                    new_resource_version = self.watch_handler(start, w, resource_version, stop)
                    resource_version = new_resource_version
                except ApiException as e:
                    if errors.is_resource_expired(e):
                        logging.info("%s: watch of %s closed with: %s" % (self.name, self.expected_type_name, str(e)))
                    elif errors.is_too_many_requests(e):
                        logging.info(
                            "%s: watch of %s returned 429 - backing off" % (self.name, self.expected_type_name))
                        time.sleep(1)
                        continue
                    else:
                        logging.warning("%s: watch of %s ended with: %s" % (self.name, self.expected_type_name, str(e)))
                except TypeError as e:
                    logging.warning(
                        "%s: watch of %s ended with unexpect error: %s" % (self.name, self.expected_type_name, str(e)))
        logging.info("list and watch end...")

    def run(self, stop: Event):
        logging.info("starting reflector %s (%s) from %s", self.expected_type_name, self.resync_period, self.name)

        def do():
            try:
                self.list_and_watch(stop)
            except Exception as e:
                if self.watch_error_handler:
                    self.watch_error_handler(self, e)
                else:
                    raise e

        wait.until_named(do, 5, stop, "reflector run")
        logging.info("stopping reflector %s (%s) from %s", self.expected_type_name, self.resync_period, self.name)


class SharedInformer(object):

    def add_event_handler(self, handler: ResourceEventHandler):
        pass

    def add_event_handler_with_resync_period(self, handler: ResourceEventHandler, resync_period: int):
        pass

    def get_store(self) -> cache.Store:
        pass

    def get_controller(self):
        pass

    def run(self, stop: Event):
        pass

    def has_synced(self) -> bool:
        pass

    def last_sync_resource_version(self) -> str:
        pass

    def set_watch_error_handler(self, handler):
        pass

    def set_transform(self, handler):
        pass


class SharedIndexInformer(SharedInformer):

    def add_indexers(self, indexers: dict):
        pass

    def get_indexer(self) -> cache.Indexer:
        pass


def _new_informer(lw: ListerWatcher, obj_type, resync_period: int,
                  h: ResourceEventHandler, client_state: cache.Store, transformer) -> Controller:
    fifo_queue = delta.DeltaFIFO(delta.DeltaFIFOOptions(
        known_objects=client_state,
        emit_delta_type_replaced=True,
    ))

    def process(obj):
        if isinstance(obj, list):
            _process_deltas(h, client_state, transformer, obj)
        else:
            raise Exception("object given as Process argument is not Deltas")

    cfg = Config(
        queue=fifo_queue,
        lister_watcher=lw,
        full_resync_period=resync_period,
        retry_on_error=False,
        process_func=process,
        obj_type=obj_type
    )
    return InnerController(cfg)


class Config(object):

    def __init__(self, queue: fifo.Queue, lister_watcher: ListerWatcher, process_func, obj_type, full_resync_period=0,
                 should_resync_func=None, retry_on_error=True, watch_error_handler=None, watch_list_page_size=100):
        self.queue = queue
        self.lister_watcher = lister_watcher
        self.process = process_func
        self.obj_type = obj_type
        self.full_resync_period = full_resync_period
        self.should_resync = should_resync_func
        self.retry_on_error = retry_on_error
        self.watch_error_handler = watch_error_handler
        self.watch_list_page_size = watch_list_page_size


class InnerController(Controller):

    def __init__(self, c: Config):
        self.config = c
        self.reflector = None
        # todo: read write lock
        self.reflector_lock = threading.RLock()

    def run(self, stop: Event):
        def quit():
            while not stop.wait():
                continue
            self.config.queue.close()
        GlobalThreadManager.new_thread(target=quit, generate_name="InnerController quit").start()
        r = Reflector("",
                      self.config.lister_watcher,
                      self.config.full_resync_period,
                      self.config.queue)
        r.should_resync = self.config.should_resync
        r.watch_list_page_size = self.config.watch_list_page_size
        if self.config.watch_error_handler:
            r.watch_error_handler = self.config.watch_error_handler
        try:
            self.reflector_lock.acquire()
            self.reflector = r
        finally:
            self.reflector_lock.release()
        t = GlobalThreadManager.new_thread(target=r.run, args=(stop,), generate_name="InnerController-start")
        t.start()
        wait.until_named(self._process_loop, 1, stop, "inner controller run")
        t.join()

    def has_synced(self) -> bool:
        return self.config.queue.has_synced()

    def last_sync_resource_version(self) -> str:
        try:
            self.reflector_lock.acquire()
            if not self.reflector:
                return ""
            return self.reflector.get_last_sync_resource_version()
        finally:
            self.reflector_lock.release()

    def _process_loop(self):
        obj = None
        while True:
            try:
                obj = self.config.queue.pop(self.config.process)
            except fifo.FIFOClosedException:
                return
            except Exception as e:
                if self.config.retry_on_error:
                    self.config.queue.add_if_not_present(obj)


def deletion_handling_meta_namespace_key_func(obj) -> str:
    if isinstance(obj, cache.DeletedFinalStateUnknown):
        return obj.key
    return cache.meta_namespace_key_func(obj)


def _process_deltas(handler: ResourceEventHandler, client_state: cache.Store, transformer, deltas: list):
    for d in deltas:
        obj = d.obj
        if transformer:
            obj = transformer(obj)
        if d.type in {delta.DeltaType.Sync, delta.DeltaType.Replaced, delta.DeltaType.Added, delta.DeltaType.Updated}:
            old, exists = client_state.get(obj)
            if exists:
                handler.on_update(old, obj)
                client_state.update(obj)
            else:
                handler.on_add(obj)
                client_state.add(obj)
        elif d.type == delta.DeltaType.Deleted:
            handler.on_delete(obj)
            client_state.delete(obj)
        else:
            logging.error("unknow delta type: %s" % d.type)


def new_informer(lw: ListerWatcher, obj_type, resync_period, h: ResourceEventHandler) -> (cache.Store, Controller):
    client_state = cache.Cache(key_func=deletion_handling_meta_namespace_key_func, indexers={})
    # todoR
    return client_state, \
           _new_informer(lw, obj_type=obj_type,
                         resync_period=resync_period, h=h, client_state=client_state, transformer=None)


class _UpdateNotification(object):

    def __init__(self, old_obj, new_obj):
        self.old_obj = old_obj
        self.new_obj = new_obj


class _AddNotification(object):

    def __init__(self, new_obj):
        self.new_obj = new_obj


class _DeleteNotification(object):

    def __init__(self, old_obj):
        self.old_obj = old_obj

class _ProcessorListener(object):

    def __init__(self, handler: ResourceEventHandler, requested_resync_period: int, resync_period: int, now: int,
                 buffer_size: int, stop_ch: threading.Event):
        self.handler = handler
        # todo: RingGrowing
        self.pending_notifications = []
        self.requested_resync_period = requested_resync_period
        self.resync_period = resync_period
        self.resync_lock = threading.Lock()
        self.next_resync = 0
        self.determine_next_resync(now)
        self.add_ch = Queue()
        self.stop_ch = stop_ch

    def determine_next_resync(self, now: int):
        with self.resync_lock:
            self.next_resync = now + self.resync_period

    def should_resync(self, now) -> bool:
        with self.resync_lock:
            if self.resync_period == 0:
                return False
            return now >= self.next_resync

    def set_resync_period(self, resync_period):
        with self.resync_lock:
            self.resync_period = resync_period

    def add(self, notification):
        self.add_ch.put_nowait(notification)

    def run(self):

        def f():
            while self.add_ch.qsize() > 0:
                try:
                    item = self.add_ch.get_nowait()
                except queue.Empty:
                    continue
                if not item:
                    continue
                if isinstance(item, _UpdateNotification):
                    self.handler.on_update(item.old_obj, item.new_obj)
                elif isinstance(item, _AddNotification):
                    self.handler.on_add(item.new_obj)
                elif isinstance(item, _DeleteNotification):
                    self.handler.on_delete(item.old_obj)
                else:
                    logging.error("unrecognized notification")

        wait.until_named(f, 1, self.stop_ch, "listener run")


def _determine_resync_period(desired, check: int) ->int:
    if desired == 0:
        return desired
    if check == 0:
        logging.warning("the specified resyncPeriod %d is invalid because this shared informer doesn't support resyncing", desired)
        return check
    if desired < check:
        logging.warning("the specified resyncPeriod %d is being increased to the minimum resyncCheckPeriod %d", desired, check)
        return check
    return desired

class _SharedProcessor(object):

    def __init__(self):
        self.listeners_started = False
        self.listeners_lock = threading.Lock()
        self.listeners = []
        self.syncing_listeners = []

    def add_listener_locked(self, listener: _ProcessorListener):
        self.listeners.append(listener)
        self.syncing_listeners.append(listener)

    def add_listener(self, listener: _ProcessorListener):
        with self.listeners_lock:
            self.add_listener_locked(listener)
            if self.listeners_started:
                wait.until_with_thread(listener.run, 1, listener.stop_ch, "listener run")

    def distribute(self, obj, sync: bool):
        if sync:
            for listener in self.syncing_listeners:
                listener.add(obj)
        else:
            for listener in self.listeners:
                listener.add(obj)

    def run(self, stop: threading.Event):
        with self.listeners_lock:
            for listener in self.listeners:
                wait.until_with_thread(listener.run, 1, listener.stop_ch, "listener run2")
            self.listeners_started = True
        # while not stop.wait(30):
        #     continue
        # with self.listeners_lock:
        #     for listener in self.listeners:
        #         listener.stop_ch.set()

    def should_resync(self) -> bool:
        with self.listeners_lock:
            self.syncing_listeners.clear()
            resync_needed = False
            now = time.time()
            for listener in self.listeners:
                if listener.should_resync(now):
                    resync_needed = True
                    self.syncing_listeners.append(listener)
                    listener.determine_next_resync(now)
            return resync_needed

    def resync_check_period_changed(self, resync_check_period: int):
        for listener in self.listeners:
            resync_period = _determine_resync_period(listener.requested_resync_period, resync_check_period)
            listener.set_resync_period(resync_period)


class _DummyController(Controller):

    def __init__(self, informer: SharedIndexInformer):
        self.informer = informer

    def run(self, stop: Event):
        pass

    def has_synced(self) -> bool:
        return self.informer.has_synced()

    def last_sync_resource_version(self) -> str:
        pass


class _SharedIndexInformer(SharedIndexInformer, ResourceEventHandler):

    def __init__(self, lw: ListerWatcher, example_obj, default_event_handler_resync_period: int, indexers: dict, stop):
        self.indexer = cache.Cache(deletion_handling_meta_namespace_key_func, indexers)
        self.controller = None
        self.processor = _SharedProcessor()
        self.cache_mutation_detector = cache.DummyMutationDetector()
        self.lister_watcher = lw
        self.obj_type = example_obj
        self.resync_check_period = default_event_handler_resync_period
        self.default_event_handler_resync_period = default_event_handler_resync_period
        self.started = False
        self.stopped = False
        self.started_lock = threading.Lock()
        self.block_deltas = threading.Lock()
        self.watch_error_handler = None
        self.transform = None
        self.stopch = stop

    def add_indexers(self, indexers: dict):
        with self.started_lock:
            if self.started:
                raise Exception("informer has already started")
            self.indexer.add_indexers(indexers)

    def get_indexer(self) -> cache.Indexer:
        return self.indexer

    def add_event_handler(self, handler: ResourceEventHandler):
        self.add_event_handler_with_resync_period(handler, self.default_event_handler_resync_period)

    def add_event_handler_with_resync_period(self, handler: ResourceEventHandler, resync_period: int):
        with self.started_lock:
            if self.stopped:
                logging.info("Handler %s was not added to shared informer because it has stopped already", str(handler))
                return
            if resync_period > 0:
                if resync_period < 1:
                    resync_period = 1
                if resync_period < self.resync_check_period:
                    if self.started:
                        logging.warning("resync_period %d is smaller than resync_check_period %d and the informer "
                                        "has already started. changing it to %d"
                                        % (resync_period, self.resync_check_period, self.resync_check_period))
                        resync_period = self.resync_check_period
                    else:
                        self.resync_check_period = resync_period
                        self.processor.resync_check_period_changed(resync_period)
            listener = _ProcessorListener(
                handler, resync_period, _determine_resync_period(resync_period, self.resync_check_period),
                int(time.time()), 1024, self.stopch)
            if not self.started:
                self.processor.add_listener(listener)
                return
            with self.block_deltas:
                self.processor.add_listener(listener)
                for item in self.indexer.list():
                    listener.add(_AddNotification(new_obj=item))

    def get_store(self) -> cache.Store:
        return self.indexer

    def get_controller(self):
        return _DummyController(self)

    def run(self, stop: Event):
        try:
            if self.has_started():
                logging.warning("the sharedIndexInformer has started, run more than once is not allowed")
                return
            fifo = delta.DeltaFIFO(delta.DeltaFIFOOptions(known_objects=self.indexer, emit_delta_type_replaced=True))
            cfg = Config(
                queue=fifo,
                lister_watcher=self.lister_watcher,
                obj_type=self.obj_type,
                full_resync_period=self.resync_check_period,
                retry_on_error=False,
                should_resync_func=self.processor.should_resync,
                process_func=self.handle_deltas,
                watch_error_handler=self.watch_error_handler
            )
            with self.started_lock:
                self.controller = InnerController(c=cfg)
                self.started = True

            t1 = GlobalThreadManager.new_thread(
                target=functools.partial(self.cache_mutation_detector.run, self.stopch),
                generate_name="cache_mutation_detector")
            t2 = GlobalThreadManager.new_thread(
                target=functools.partial(self.processor.run, self.stopch),
                generate_name="processor.run")
            t1.start()
            t2.start()
            self.controller.run(self.stopch)
            t1.join()
            t2.join()
            with self.started_lock:
                self.stopped = True

        except Exception as e:
            logging.error("informer run error: %s", str(e))

    def has_synced(self) -> bool:
        with self.started_lock:
            return self.controller.has_synced()

    def last_sync_resource_version(self) -> str:
        with self.started_lock:
            if not self.controller:
                return ""
            return self.controller.last_sync_resource_version()

    def set_watch_error_handler(self, handler):
        with self.started_lock:
            if self.started:
                raise Exception("informer has already started")
            self.watch_error_handler = handler

    def set_transform(self, handler):
        with self.started_lock:
            if self.started:
                raise Exception("informer has already started")
            self.transform = handler

    def has_started(self) -> bool:
        with self.started_lock:
            return self.started

    def handle_deltas(self, obj):
        with self.block_deltas:
            if isinstance(obj, list):
                _process_deltas(self, self.indexer, self.transform, obj)
                return
            raise Exception("object given as Process argument is not Deltas")

    def on_add(self, obj):
        self.cache_mutation_detector.add_object(obj)
        self.processor.distribute(_AddNotification(new_obj=obj), False)

    def on_update(self, old_obj, new_obj):
        is_sync = False
        if hasattr(new_obj, "metadata") and hasattr(new_obj.metadata, "resource_version"):
            is_sync = new_obj.metadata.resource_version == old_obj.metadata.resource_version
        self.cache_mutation_detector.add_object(new_obj)
        self.processor.distribute(_UpdateNotification(new_obj=new_obj, old_obj=old_obj), is_sync)

    def on_delete(self, obj):
        self.processor.distribute(_DeleteNotification(old_obj=obj), False)


class ListWatch(ListerWatcher):

    def __init__(self, list_func, watch_func, disable_chunking=False):
        self.list_func = list_func
        self.watch_func = watch_func
        self.disable_chunking = disable_chunking

    def list(self, options: meta_v1.ListOptions):
        return self.list_func(options)

    def watch(self, options: meta_v1.ListOptions) -> watch.Interface:
        return self.watch_func(options)



import time
import unittest
from kubernetes import config
from kubernetes.client.models.core_v1_event import CoreV1Event
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from kubernetes.client.models.v1_object_reference import V1ObjectReference
from dragonk8s.lib.client.kubernetes.typed.core.v1.event_expansion import EventSinkImpl
from kubernetes.client.exceptions import ApiException
from dragonk8s.lib.apimachinery.pkg.api import errors
import datetime
from kubernetes.client.models.v1_status import V1Status
from dragonk8s.lib.client.util.trans import parse_v1_status, parse_json_to_object, parse_json_to_object_by_class_name


class EventSinkImplTest(unittest.TestCase):

    def setUp(self):
        config.load_kube_config(config_file="/Users/bytedance/.kube/config")
        self.event_sink = EventSinkImpl()

    def test_create_event(self):
        event = CoreV1Event(
            api_version="v1",
            kind="Event",
            metadata=V1ObjectMeta(
                name="m13",
                namespace="kube-system"
            ),
            # event_time="2022-11-30T02:21:25.458481Z",
            action="Action11",
            type="Normal",
            reason="somereason",
            message="some message",
            involved_object=V1ObjectReference(
                api_version="v1",
                kind="Pod",
                name="coredns-7f74c56694-qsrts",
                namespace='kube-system'
            ),
            reporting_component="default-scheduler11",
            reporting_instance="default-scheduler-vm10-0-11-1211111",
            count=2,
        )
        try:
            self.event_sink.create(event)
        except ApiException as e:
            print(e.status)
            print(e.reason)
            # status = parse_v1_status(e.body)
            # status = parse_json_to_object(e.body, V1Status())
            status = parse_json_to_object_by_class_name(e.body, "V1Status")
            print(status.to_dict())
            print(status.api_version)
            print(status.code)
            print(status.reason)
            print(status.details)
            raise e

    def test_create_event_exist(self):
        event = CoreV1Event(
            api_version="v1",
            kind="Event",
            metadata=V1ObjectMeta(
                name="m12",
                namespace="kube-system"
            ),
            event_time="2022-11-30T02:21:25.458481Z",
            action="Action11",
            type="Normal",
            reason="somereason",
            message="some message",
            involved_object=V1ObjectReference(
                api_version="v1",
                kind="Pod",
                name="coredns-7f74c56694-qsrts",
                namespace='kube-system'
            ),
            reporting_component="default-scheduler11",
            reporting_instance="default-scheduler-vm10-0-11-1211111",
            count=2,
            first_timestamp="2022-11-30T02:21:25.458481Z",
        )
        try:
            self.event_sink.create(event)
        except ApiException as e:
            if errors.is_already_exists(e):
                print("is exist")
            else:
                raise e

    def test_update_event(self):
        event = CoreV1Event(
            api_version="v1",
            kind="Event",
            metadata=V1ObjectMeta(
                name="m1",
                namespace="kube-system"
            ),
            event_time="2022-11-30T02:21:25.458481Z",
            action="Action11",
            type="Normal",
            reason="somereason111",
            message="some message",
            involved_object=V1ObjectReference(
                api_version="v1",
                kind="Pod",
                name="coredns-7f74c56694-qsrts",
                namespace='kube-system'
            ),
            reporting_component="default-scheduler11",
            reporting_instance="default-scheduler-vm10-0-11-1211111"
        )
        self.event_sink.update(event)

    def test_patch_event(self):
        event = self.event_sink.client.read_namespaced_event("m12", "kube-system")
        print(event.to_dict())
        tot = datetime.datetime.fromtimestamp(time.time()).strftime("%Y-%m-%dT%H:%M:%SZ")
        print(tot)
        data = {
            "reportingComponent": "wahah222",
            "lastTimestamp": datetime.datetime.now(),
            "count": 22,
        }
        event = self.event_sink.patch(event, data)
        print(event.to_dict())

    def test_not_found(self):
        try:
            event = self.event_sink.client.read_namespaced_event("m2222", "kube-system")
            print(event.to_dict())
        except ApiException as e:
            print(e.status)

    def test_get(self):
        try:
            event = self.event_sink.client.read_namespaced_event("m12", "kube-system")
            print(event.to_dict())
            print(event.first_timestamp.timestamp())
            print(datetime.datetime.fromtimestamp(time.time()))
        except ApiException as e:
            print(e.status)


from __future__ import annotations

import copy
import json
import logging
import threading
import time

from dragonk8s.lib.client.informers.informer import BaseInformer
from dragonk8s.lib.client.informers.dragon.replicaset import ReplicaSetInformer
from dragonk8s.lib.client.tools import record
from kubernetes.client.models.v1_event_source import V1EventSource
from dragonk8s.lib.client.kubernetes.typed.core.v1.event_expansion import EventSinkImpl
from dragonk8s.pkg.controller import controller_utils
from dragonk8s.lib.client.util.workqueue import RateLimitingType
from dragonk8s.dragon.models import IoDragonAppsV1ReplicaSet
from kubernetes.client.models import V1Pod, V1OwnerReference
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.client.tools import informer
from dragonk8s.lib.apimachinery.pkg.util import wait
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.client.tools import cache
from dragonk8s.lib.apimachinery.pkg.api import errors
from dragonk8s.lib.apimachinery.pkg import labels
from dragonk8s.pkg.controller.replicaset import replica_set_utils
from dragonk8s.lib.util import timeutil
from dragonk8s.pkg.api.v1 import pod_util
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager

controllerUIDIndex = "controllerUID"


def _get_pod_to_delete(filtered_pods: list, related_pods: list, diff) -> list:
    # todo getPodsRankedByRelatedPodsOnSameNode
    if diff < len(filtered_pods):
        return filtered_pods[:diff]
    logging.warning("diff greater than filtered_pods??")
    return filtered_pods


def _get_pod_keys(pods: list) -> list:
    pod_keys = []
    for pod in pods:
        pod_keys.append(controller_utils.pod_key(pod))
    return pod_keys


class ReplicaSetController(object):

    def __init__(self, rs_informer: ReplicaSetInformer, pod_informer: BaseInformer, kube_client, burst_replicas: int,
                 stop: threading.Event):
        event_broadcaster = record.EventBroadcasterImpl(stop=stop)
        event_broadcaster.start_recording_to_sink(EventSinkImpl())
        event_broadcaster.start_structured_logging()
        # todo: metrics

        self.kube_client = apigvk.get_resource_client(kube_client, apigvk.IoDragonAppsV1ReplicaSet)
        self.pod_control = controller_utils.RealPodControl(
            kube_client=kube_client,
            recorder=event_broadcaster.new_recorder(V1EventSource(component="dragon-replicaset-controller")))
        self.burst_replicas = burst_replicas
        self.expectations = controller_utils.UIDTrackingControllerExpectations(controller_utils.ControllerExpectations())
        self.queue = RateLimitingType()
        self._type = apigvk.IoDragonAppsV1ReplicaSet
        self.stop = stop

        rs_informer.informer().add_event_handler(informer.ResourceEventHandlerFuncs(
            add_func=self._add_rs,
            update_func=self._update_rs,
            delete_func=self._delete_rs,
        ))

        def uid_index(obj) -> list:
            controller_ref = meta_v1.get_controller_of(obj)
            if controller_ref is None:
                return []
            return [controller_ref.uid]
        rs_informer.informer().add_indexers({
            controllerUIDIndex: uid_index,
        })
        self.rs_lister = rs_informer.lister()
        self.rs_indexer = rs_informer.informer().get_indexer()
        self.rs_lister_synced = rs_informer.informer().has_synced

        pod_informer.informer().add_event_handler(informer.ResourceEventHandlerFuncs(
            add_func=self._add_pod,
            update_func=self._update_pod,
            delete_func=self._delete_pod,
        ))
        self.pod_lister = pod_informer.lister()
        self.pod_lister_synced = pod_informer.informer().has_synced
        self.sync_handler = self._sync_replicaset

    def _sync_replicaset(self, key: str):
        logging.info("sync replicaset %s" % key)
        try:
            return self.__sync_replicaset(key)
        finally:
            logging.info("sync replicaset %s done..." % key)

    def _enqueue_rs(self, rs: IoDragonAppsV1ReplicaSet):
        try:
            key = informer.deletion_handling_meta_namespace_key_func(rs)
        except Exception as e:
            logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.to_str(), str(e)))
            return
        self.queue.add(key)

    def _enqueue_rs_after(self, rs: IoDragonAppsV1ReplicaSet, duration: int):
        try:
            key = informer.deletion_handling_meta_namespace_key_func(rs)
        except Exception as e:
            logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.to_str(), str(e)))
            return
        self.queue.add_after(key, duration)

    def _add_rs(self, obj: IoDragonAppsV1ReplicaSet):
        rs = obj
        logging.info("adding %s %s/%s" % (self._type.kind, rs.metadata.namespace, rs.metadata.name))
        self._enqueue_rs(rs)

    def _update_rs(self, old_obj: IoDragonAppsV1ReplicaSet, new_obj: IoDragonAppsV1ReplicaSet):
        if new_obj.metadata.uid != old_obj.metadata.uid:
            try:
                key = informer.deletion_handling_meta_namespace_key_func(old_obj)
            except Exception as e:
                logging.error("couldn't get key for %s %s: %s" % (self._type.kind, old_obj.to_str(), str(e)))
                return
            self._delete_rs(cache.DeletedFinalStateUnknown(
                key=key,
                obj=old_obj,
            ))
        if old_obj.spec.replicas != new_obj.spec.replicas:
            logging.info("%s %s updated. desired pod count change: %d->%d"
                         % (self._type.kind, new_obj.metadata.name, old_obj.spec.replicas, new_obj.spec.replicas))
        self._enqueue_rs(new_obj)

    def _delete_rs(self, obj: IoDragonAppsV1ReplicaSet|cache.DeletedFinalStateUnknown):
        if not isinstance(obj, IoDragonAppsV1ReplicaSet):
            if not isinstance(obj, cache.DeletedFinalStateUnknown):
                logging.error("couldn't get object from tombstone %s" % (str(obj)))
                return
            rs = obj.obj
        else:
            rs = obj
        try:
            key = informer.deletion_handling_meta_namespace_key_func(rs)
        except Exception as e:
            logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.to_str(), str(e)))
            return
        logging.info("Deleting %s %s" % (self._type.kind, key))
        self.expectations.delete_expectations(key)
        self.queue.add(key)

    def _get_pod_replicasets(self, pod: V1Pod) -> list:
        try:
            rss = self.rs_lister.get_pod_replica_sets(pod)
        except Exception as e:
            return []
        if len(rss) > 1:
            logging.error("user error! more than one %s is selecting pods with labels: %s"
                          % (self._type.kind, pod.metadata.labels))
        return rss

    def _add_pod(self, obj: V1Pod):
        pod = obj
        if pod.metadata.deletion_timestamp:
            self._delete_pod(pod)
            return
        controller_ref = meta_v1.get_controller_of(pod)
        if controller_ref:
            rs = self._resolve_controller_ref(pod.metadata.namespace, controller_ref)
            if rs is None:
                return
            try:
                key = informer.deletion_handling_meta_namespace_key_func(rs)
            except Exception as e:
                logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.to_str(), str(e)))
                return
            logging.info("Pod %s created: %s" % (pod.metadata.name, pod.metadata.name))
            self.expectations.ce.creation_observed(key)
            self.queue.add(key)
            return
        rss = self._get_pod_replicasets(pod)
        if len(rss) == 0:
            return
        logging.info("Orphan Pod %s created: %s" % (pod.metadata.name, pod.to_str()))
        for rs in rss:
            self._enqueue_rs(rs)

    def _update_pod(self, old_obj: V1Pod, new_obj: V1Pod):
        if old_obj.metadata.resource_version == new_obj.metadata.resource_version:
            return
        label_changed = new_obj.metadata.labels != old_obj.metadata.labels
        if new_obj.metadata.deletion_timestamp:
            self._delete_pod(new_obj)
            if label_changed:
                self._delete_pod(old_obj)
            return
        new_controller_ref = meta_v1.get_controller_of(new_obj)
        old_controller_ref = meta_v1.get_controller_of(old_obj)
        controller_ref_changed = new_controller_ref.to_dict() != old_controller_ref.to_dict()
        if controller_ref_changed and old_controller_ref:
            rs = self._resolve_controller_ref(old_obj.metadata.namespace, old_controller_ref)
            if rs:
                self._enqueue_rs(rs)
        if new_controller_ref:
            rs = self._resolve_controller_ref(new_obj.metadata.namespace, new_controller_ref)
            if rs is None:
                return
            # logging.info("pod %s updated,  %s -> %s" %
            #              (new_obj.metadata.name, old_obj.metadata.to_str(), new_obj.metadata.to_str()))
            logging.info("pod %s updated" % new_obj.metadata.name)
            self._enqueue_rs(rs)
            if not pod_util.is_pod_ready(old_obj) and pod_util.is_pod_ready(new_obj) and rs.spec.min_ready_seconds > 0:
                logging.info("%s %s will be enqueued after %ds for availability check"
                             % (self._type.kind, rs.metadata.name, rs.spec.min_ready_seconds))
                self._enqueue_rs_after(rs, rs.spec.min_ready_seconds+1)
            return
        if label_changed or controller_ref_changed:
            rss = self._get_pod_replicasets(new_obj)
            if len(rss) == 0:
                return
            # logging.info("Orphan Pod %s updated, objectMeta %s -> %s"
            #              % (new_obj.metadata.name, old_obj.metadata.to_str(), new_obj.metadata.to_str()))
            logging.info("Orphan Pod %s updated " % new_obj.metadata.name)
            for rs in rss:
                self._enqueue_rs(rs)

    def _delete_pod(self, obj: V1Pod|cache.DeletedFinalStateUnknown):
        if not isinstance(obj, V1Pod):
            if not isinstance(obj, cache.DeletedFinalStateUnknown):
                logging.error("couldn't get object from tombstone %s" % str(obj))
                return
            pod = obj.obj
        else:
            pod = obj
        controller_ref = meta_v1.get_controller_of(pod)
        if not controller_ref:
            return
        rs = self._resolve_controller_ref(pod.metadata.namespace, controller_ref)
        if not rs:
            return
        try:
            key = informer.deletion_handling_meta_namespace_key_func(rs)
        except Exception as e:
            logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.to_str(), str(e)))
            return
        logging.info("pod %s/%s deleted throuth %s, timestamp: %s"
                     % (pod.metadata.name, pod.metadata.name, "dragon-replicaset-controller",
                        timeutil.to_time_str_with_ns(pod.metadata.deletion_timestamp)))
        self.expectations.deletion_observed(key, controller_utils.pod_key(pod))
        self.queue.add(key)

    def _resolve_controller_ref(self, namespace: str, controller_ref: V1OwnerReference) -> IoDragonAppsV1ReplicaSet|None:
        if controller_ref.kind != self._type.kind:
            return None
        try:
            rs = self.rs_lister.with_namespace(namespace).get(controller_ref.name)
        except Exception as e:
            if isinstance(e, errors.ApiException) and errors.is_not_found(e):
                return None
            logging.error("get rs %s error: %s" % (controller_ref.name, str(e)))
            return None
        if rs.metadata.uid != controller_ref.uid:
            return None
        return rs

    def _process_next_work_item(self) -> bool:
        if self.stop.is_set():
            return False
        key, quit = self.queue.get()
        if quit:
            return False
        try:
            self.sync_handler(key)
            # todo
        except Exception as e:
            logging.error("sync %s error: %s" % (key, str(e)))
            self.queue.add_rate_limited(key)
            return True
        else:
            self.queue.forget(key)
            return True
        finally:
            self.queue.done(key)

    def _worker(self):
        while self._process_next_work_item():
            pass
        logging.debug("worker quit")

    def run(self, workers: int):

        def quit():
            while not self.stop.wait():
                continue
            self.queue.shutdown()
        GlobalThreadManager.new_thread(target=quit, generate_name="controller quit").start()
        try:
            controller_name = self._type.kind.lower()
            logging.info("Starting %s controller" % controller_name)
            if not cache.wait_for_named_cache_sync(controller_name, self.stop, self.pod_lister_synced, self.rs_lister_synced):
                return
            for i in range(workers):
                wait.until_with_thread(self._worker, 1, self.stop, "worker-%d" % i)
            logging.info("controller wait to done...")
            self.stop.wait()
            logging.info("Stopping %s controller" % controller_name)
        except Exception as e:
            logging.info("controller run error: %s" % str(e))

    def _get_replicasets_with_same_controller(self, rs: IoDragonAppsV1ReplicaSet) -> list:
        controller_ref = meta_v1.get_controller_of(rs)
        if not controller_ref:
            logging.warning("ReplicaSet has no controller: %s" % rs.to_str())
            return []
        try:
            objs = self.rs_indexer.by_index(controllerUIDIndex, controller_ref.uid)
        except Exception as e:
            logging.error(" index error: %s" % str(e))
            return []
        return objs

    def _get_indirectly_related_pods(self, rs: IoDragonAppsV1ReplicaSet) -> list:
        related_pods = []
        seen = {}
        for related_rs in self._get_replicasets_with_same_controller(rs):
            try:
                selector = meta_v1.labels_selector_as_selector(related_rs.spec.selector)
            except Exception as e:
                logging.error("_get_indirectly_related_pods labels_selector_as_selector error: %s" % str(e))
                continue
            try:
                pods = self.pod_lister.with_namespace(related_rs.metadata.namespace).list(selector)
            except Exception as e:
                logging.error("_get_indirectly_related_pods, list pod error: %s" % str(e))
                return []
            for pod in pods:
                if pod.metadata.uid in seen:
                    other_rs = seen[pod.metadata.uid]
                    logging.info("pod %s/%s is owned by both %s %s/%s and %s %s/%s"
                                 % (pod.metadata.namespace, pod.metadata.name, self._type.kind,
                                    other_rs.metadata.namespace, other_rs.metadata.name, self._type.kind,
                                    related_rs.metadata.namespace, related_rs.metadata.name))
                    continue
                seen[pod.metadata.uid] = related_rs
                related_pods.append(pod)
            logging.info("Found related pods, count %d" % len(related_pods))
            return related_pods

    def _manage_replicas(self, filtered_pods: list, rs: IoDragonAppsV1ReplicaSet):
        logging.info("manage replicas for rs %s/%s" % (rs.metadata.namespace, rs.metadata.name))
        diff = len(filtered_pods) - rs.spec.replicas
        try:
            # cache.DeletionHandlingMetaNamespaceKeyFunc
            rs_key = informer.deletion_handling_meta_namespace_key_func(rs)
        except Exception as e:
            logging.error("couldn't get key for %s %s: %s" % (self._type.kind, rs.metadata.name, str(e)))
            return
        if diff < 0:
            diff = diff * -1
            if diff > self.burst_replicas:
                diff = self.burst_replicas
            self.expectations.ce.expect_creations(rs_key, diff)
            logging.info("Too few replicas, need: %d, creating: %d", rs.spec.replicas, diff)
            # todo: 并行
            success_creations = 0
            for i in range(diff):
                try:
                    self.pod_control.create_pod(
                        rs.metadata.namespace, rs.spec.template, rs, meta_v1.new_controller_ref(rs, self._type))
                except errors.ApiException as e:
                    if errors.has_status_cause(e, errors.NamespaceTerminatingCause):
                        return
                else:
                    success_creations += 1
            skipped_pods = diff - success_creations
            if skipped_pods > 0:
                logging.info("create pods failure, Skipping creation of %d pods, decrementing expectations for %s %s/%s"
                             % (skipped_pods, self._type.kind, rs.metadata.namespace, rs.metadata.name))
                for i in range(skipped_pods):
                    self.expectations.ce.creation_observed(rs_key)
            return
        elif diff > 0:
            if diff > self.burst_replicas:
                diff = self.burst_replicas
            logging.info("Too many replicas, need: %d, deleting: %d", rs.spec.replicas, diff)
            # related_pods = self._get_indirectly_related_pods(rs)
            pods_to_delete = _get_pod_to_delete(filtered_pods, [], diff)
            self.expectations.expect_deletions(rs_key, _get_pod_keys(pods_to_delete))
            # todo 并行
            for pod in pods_to_delete:
                try:
                    self.pod_control.delete_pod(rs.metadata.namespace, pod.metadata.name, rs)
                except Exception as e:
                    pod_key = controller_utils.pod_key(pod)
                    self.expectations.deletion_observed(rs_key, pod_key)
                    if not isinstance(e, errors.ApiException) or not errors.is_not_found(e):
                        logging.warning("failed to delete %s, decremented expectations for %s %s/%s"
                                        % (pod_key, self._type.kind, rs.metadata.namespace, rs.metadata.name))
        else:
            logging.info("no diff replicas for rs rs %s/%s" % (rs.metadata.namespace, rs.metadata.name))

    def _claim_pods(self, rs: IoDragonAppsV1ReplicaSet, selector: labels.Selector, filtered_pods: list) -> list:

        def get_rs():
            fresh = self.kube_client.get(name=rs.metadata.name, namespace=rs.metadata.namespace)
            if fresh.metadata.uid != rs.metadata.uid:
                raise Exception("original %s %s/%s is gone: got uid: %s, wanted %s"
                                % (self._type.kind, rs.metadata.namespace, rs.metadata.name,
                                   fresh.metadata.uid, rs.metadata.uid))
            return fresh
        cm = controller_utils.PodControllerRefManager(self.pod_control,
                                                      rs, selector, self._type,
                                                      controller_utils.recheck_deletion_timestamp(get_rs))
        return cm.claim_pods(filtered_pods)

    def __sync_replicaset(self, key: str):
        namespace, name = cache.split_meta_namespace_key(key)
        rs = None
        try:
            rs = self.rs_lister.with_namespace(namespace).get(name)
        except errors.ApiException as e:
            if errors.is_not_found(e):
                logging.info("%s %s has been deleted" % (self._type.kind, key))
                self.expectations.delete_expectations(key)
                return
        if rs is None:
            logging.error("unexpect error, rs is None")
            return
        rs_need_sync = self.expectations.ce.satisfied_expectations(key)
        try:
            selector = meta_v1.labels_selector_as_selector(rs.spec.selector)
        except IndexError as e:
            logging.error("error converting pod selector to selector for rs %s/%s: %s" % (namespace, name, str(e)))
            return
        all_pods = self.pod_lister.with_namespace(rs.metadata.namespace).list(labels.every_thing)
        filtered_pods = controller_utils.filter_active_pods(all_pods)
        filtered_pods = self._claim_pods(rs, selector, filtered_pods)
        logging.info("all pods: %d, filtered_pods: %d" % (len(all_pods), len(filtered_pods)))
        manage_replica_exception = None
        if rs_need_sync and not rs.metadata.deletion_timestamp:
            try:
                self._manage_replicas(filtered_pods, rs)
            except Exception as e:
                manage_replica_exception = e
        if not rs_need_sync:
            logging.info("rs %s/%s is not needed to sync replicas" % (rs.metadata.namespace, rs.metadata.name))
        rs = copy.deepcopy(rs)
        new_status = replica_set_utils.calculate_status(rs, filtered_pods, manage_replica_exception)
        updated_rs = replica_set_utils.update_replicaset_status(self.kube_client, rs, new_status)
        if manage_replica_exception is None and updated_rs.spec.min_ready_seconds > 0 and \
            updated_rs.status.ready_replicas == updated_rs.spec.replicas and \
            updated_rs.status.available_replicas != updated_rs.spec.replicas:
            self.queue.add_after(key, updated_rs.spec.min_ready_seconds)
        if manage_replica_exception is not None:
            raise manage_replica_exception

    def shut_down(self):
        logging.info("replicaset-controller quit...")
        self.queue.shutdown()



from __future__ import annotations

import time

from kubernetes.client.models import V1PodCondition, V1PodStatus, V1Pod
from dragonk8s.pkg.api import types

class PodConditionType(object):
    ContainersReady = "ContainersReady"
    PodInitialized = "Initialized"
    PodReady = "Ready"
    PodScheduled = "PodScheduled"


def get_pod_condition_from_list(conditions: list, condition_type: str) -> (int, V1PodCondition|None):
    if not conditions:
        return -1, None
    for i in range(len(conditions)):
        condition = conditions[i]
        if condition.type == condition_type:
            return i, condition
    return -1, None


def get_pod_condition(status: V1PodStatus, condition_type: str) -> (int, V1PodCondition|None):
    if not status:
        return -1, None
    return get_pod_condition_from_list(status.conditions, condition_type)


def get_pod_ready_condition(status: V1PodStatus) -> V1PodCondition|None:
    _, condition = get_pod_condition(status, PodConditionType.PodReady)
    return condition


def is_pod_ready_condition_true(status: V1PodStatus) -> bool:
    condition = get_pod_ready_condition(status)
    return condition is not None and condition.status == types.ConditionTrue


def is_pod_ready(pod: V1Pod) -> bool:
    return is_pod_ready_condition_true(pod.status)


def is_pod_available(pod: V1Pod, min_ready_seconds: int, now: float) -> bool:
    if not is_pod_ready(pod):
        return False
    c = get_pod_ready_condition(pod.status)
    if not c.last_transition_time:
        last_transition_time = 0
    else:
        last_transition_time = c.last_transition_time.timestamp()
    if min_ready_seconds == 0 or (last_transition_time != 0 and last_transition_time+min_ready_seconds < now):
        return True
    return False



from threading import Lock, Condition
from dragonk8s.lib.client.tools import cache


class FIFOClosedException(Exception):
    pass


class RequeueException(Exception):

    def __init__(self, e: Exception):
        self.e = e
        super(RequeueException, self).__init__()


class Queue(cache.Store):

    def pop(self, pop_process_func):
        pass

    def add_if_not_present(self, obj):
        pass

    def has_synced(self) -> bool:
        pass

    def close(self):
        pass


def pop_process_func(obj):
    return obj


def pop(queue: Queue):
    return queue.pop(pop_process_func)


class FIFO(Queue):

    def __init__(self, key_func):
        self._lock = Lock()
        self._items = {}
        self._queue = []
        self._populated = False
        self._initial_population_count = 0
        self._key_func = key_func
        self._closed = False
        self._cond = Condition()

    def add(self, obj):
        _id = self._key_func(obj)
        try:
            self._lock.acquire(True)
            self._populated = True
            if _id not in self._items:
                self._queue.append(_id)
            self._items[_id] = obj
            self._cond.notify_all()
        finally:
            self._lock.release()

    def update(self, obj):
        self.add(obj)

    def delete(self, obj):
        _id = self._key_func(obj)
        try:
            self._lock.acquire(True)
            self._populated = True
            if _id in self._items:
                del self._items[_id]
        finally:
            self._lock.release()

    def list(self):
        try:
            self._lock.acquire(True)
            rlist = []
            for k, item in self._items:
                rlist.append(item)
            return rlist
        finally:
            self._lock.release()

    def list_keys(self):
        try:
            self._lock.acquire(True)
            rlist = []
            for k, item in self._items:
                rlist.append(k)
            return rlist
        finally:
            self._lock.release()

    def get(self, obj):
        key = self._key_func(obj)
        return self.get_by_key(key)

    def get_by_key(self, key):
        try:
            self._lock.acquire(True)
            if key not in self._items:
                return None, False
            return self._items[key], True
        finally:
            self._lock.release()

    def is_closed(self) -> bool:
        try:
            self._lock.acquire(True)
            return self._closed
        finally:
            self._lock.release()

    def replace(self, data, resource_version):
        items = {}
        for item in data:
            key = self._key_func(item)
            items[key] = item
        try:
            self._lock.acquire(True)
            if not self._populated:
                self._populated = True
                self._initial_population_count = len(items)
            self._items = items
            self._queue = self._queue[:]
            for key, item in items.items():
                self._queue.append(key)
            if len(self._queue) > 0:
                self._cond.notify_all()
        finally:
            self._lock.release()

    def resync(self):
        try:
            self._lock.acquire(True)
            in_queue = set()
            for key in self._queue:
                in_queue.add(key)
            for key, item in self._items.items():
                if key not in in_queue:
                    self._queue.append(key)
            if len(self._queue) > 0:
                self._cond.notify_all()
        finally:
            self._lock.release()

    def pop(self, pop_process_func):
        try:
            self._lock.acquire(True)
            while True:
                while len(self._queue) == 0:
                    if self._closed:
                        raise FIFOClosedException()
                    self._cond.wait()
                _id = self._queue.pop(0)
                if self._initial_population_count > 0:
                    self._initial_population_count -= 1
                if _id not in self._items:
                    continue
                item = self._items[_id]
                del self._items[_id]
                try:
                    pop_process_func(item)
                except RequeueException as e:
                    self._add_if_not_presetn(_id, item)
                    raise e
                return item
        finally:
            self._lock.release()

    def _add_if_not_presetn(self, _id, obj):
        self._populated = True
        if _id in self._items:
            return
        self._queue.append(_id)
        self._items[_id] = obj
        self._cond.notify_all()

    def add_if_not_present(self, obj):
        _id = self._key_func(obj)
        try:
            self._lock.acquire(True)
            self._add_if_not_presetn(_id, obj)
        finally:
            self._lock.release()

    def has_synced(self) -> bool:
        try:
            self._lock.acquire(True)
            return self._populated and self._initial_population_count == 0
        finally:
            self._lock.release()

    def close(self):
        try:
            self._lock.acquire(True)
            self._closed = True
            self._cond.notify_all()
        finally:
            self._lock.release()



# coding: utf-8

# flake8: noqa
"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1.24.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from dragonk8s.dragon.models.io_dragon_apps_v1_job import IoDragonAppsV1Job
from dragonk8s.dragon.models.io_dragon_apps_v1_job_list import IoDragonAppsV1JobList
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec import IoDragonAppsV1JobSpec
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_selector import IoDragonAppsV1JobSpecSelector
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_selector_match_expressions import IoDragonAppsV1JobSpecSelectorMatchExpressions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template import IoDragonAppsV1JobSpecTemplate
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec import IoDragonAppsV1JobSpecTemplateSpec
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity import IoDragonAppsV1JobSpecTemplateSpecAffinity
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinity
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity_preference import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinityPreference
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity_preference_match_expressions import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinityPreferenceMatchExpressions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity_preferred_during_scheduling_ignored_during_execution import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinityPreferredDuringSchedulingIgnoredDuringExecution
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity_required_during_scheduling_ignored_during_execution import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinityRequiredDuringSchedulingIgnoredDuringExecution
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_node_affinity_required_during_scheduling_ignored_during_execution_node_selector_terms import IoDragonAppsV1JobSpecTemplateSpecAffinityNodeAffinityRequiredDuringSchedulingIgnoredDuringExecutionNodeSelectorTerms
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinity
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity_pod_affinity_term import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinityPodAffinityTerm
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity_pod_affinity_term_label_selector import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinityPodAffinityTermLabelSelector
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity_pod_affinity_term_namespace_selector import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinityPodAffinityTermNamespaceSelector
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity_preferred_during_scheduling_ignored_during_execution import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinityPreferredDuringSchedulingIgnoredDuringExecution
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_affinity_required_during_scheduling_ignored_during_execution import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAffinityRequiredDuringSchedulingIgnoredDuringExecution
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_affinity_pod_anti_affinity import IoDragonAppsV1JobSpecTemplateSpecAffinityPodAntiAffinity
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_aws_elastic_block_store import IoDragonAppsV1JobSpecTemplateSpecAwsElasticBlockStore
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_azure_disk import IoDragonAppsV1JobSpecTemplateSpecAzureDisk
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_azure_file import IoDragonAppsV1JobSpecTemplateSpecAzureFile
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_cephfs import IoDragonAppsV1JobSpecTemplateSpecCephfs
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_cephfs_secret_ref import IoDragonAppsV1JobSpecTemplateSpecCephfsSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_cinder import IoDragonAppsV1JobSpecTemplateSpecCinder
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_cinder_secret_ref import IoDragonAppsV1JobSpecTemplateSpecCinderSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_config_map import IoDragonAppsV1JobSpecTemplateSpecConfigMap
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_config_map_items import IoDragonAppsV1JobSpecTemplateSpecConfigMapItems
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_config_map_ref import IoDragonAppsV1JobSpecTemplateSpecConfigMapRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_containers import IoDragonAppsV1JobSpecTemplateSpecContainers
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_csi import IoDragonAppsV1JobSpecTemplateSpecCsi
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_csi_node_publish_secret_ref import IoDragonAppsV1JobSpecTemplateSpecCsiNodePublishSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_dns_config import IoDragonAppsV1JobSpecTemplateSpecDnsConfig
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_dns_config_options import IoDragonAppsV1JobSpecTemplateSpecDnsConfigOptions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_downward_api import IoDragonAppsV1JobSpecTemplateSpecDownwardAPI
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_downward_api_field_ref import IoDragonAppsV1JobSpecTemplateSpecDownwardAPIFieldRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_downward_api_items import IoDragonAppsV1JobSpecTemplateSpecDownwardAPIItems
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_downward_api_resource_field_ref import IoDragonAppsV1JobSpecTemplateSpecDownwardAPIResourceFieldRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_empty_dir import IoDragonAppsV1JobSpecTemplateSpecEmptyDir
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_env import IoDragonAppsV1JobSpecTemplateSpecEnv
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_env_from import IoDragonAppsV1JobSpecTemplateSpecEnvFrom
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral import IoDragonAppsV1JobSpecTemplateSpecEphemeral
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_containers import IoDragonAppsV1JobSpecTemplateSpecEphemeralContainers
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplate
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template_spec import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplateSpec
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template_spec_data_source import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplateSpecDataSource
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template_spec_data_source_ref import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplateSpecDataSourceRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template_spec_resources import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplateSpecResources
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ephemeral_volume_claim_template_spec_selector import IoDragonAppsV1JobSpecTemplateSpecEphemeralVolumeClaimTemplateSpecSelector
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_fc import IoDragonAppsV1JobSpecTemplateSpecFc
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_flex_volume import IoDragonAppsV1JobSpecTemplateSpecFlexVolume
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_flex_volume_secret_ref import IoDragonAppsV1JobSpecTemplateSpecFlexVolumeSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_flocker import IoDragonAppsV1JobSpecTemplateSpecFlocker
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_gce_persistent_disk import IoDragonAppsV1JobSpecTemplateSpecGcePersistentDisk
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_git_repo import IoDragonAppsV1JobSpecTemplateSpecGitRepo
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_glusterfs import IoDragonAppsV1JobSpecTemplateSpecGlusterfs
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_host_aliases import IoDragonAppsV1JobSpecTemplateSpecHostAliases
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_host_path import IoDragonAppsV1JobSpecTemplateSpecHostPath
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_image_pull_secrets import IoDragonAppsV1JobSpecTemplateSpecImagePullSecrets
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_iscsi import IoDragonAppsV1JobSpecTemplateSpecIscsi
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_iscsi_secret_ref import IoDragonAppsV1JobSpecTemplateSpecIscsiSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_label_selector import IoDragonAppsV1JobSpecTemplateSpecLabelSelector
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle import IoDragonAppsV1JobSpecTemplateSpecLifecycle
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle1 import IoDragonAppsV1JobSpecTemplateSpecLifecycle1
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_post_start import IoDragonAppsV1JobSpecTemplateSpecLifecyclePostStart
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_post_start_exec import IoDragonAppsV1JobSpecTemplateSpecLifecyclePostStartExec
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_post_start_http_get import IoDragonAppsV1JobSpecTemplateSpecLifecyclePostStartHttpGet
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_post_start_http_get_http_headers import IoDragonAppsV1JobSpecTemplateSpecLifecyclePostStartHttpGetHttpHeaders
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_post_start_tcp_socket import IoDragonAppsV1JobSpecTemplateSpecLifecyclePostStartTcpSocket
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_lifecycle_pre_stop import IoDragonAppsV1JobSpecTemplateSpecLifecyclePreStop
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_liveness_probe import IoDragonAppsV1JobSpecTemplateSpecLivenessProbe
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_liveness_probe1 import IoDragonAppsV1JobSpecTemplateSpecLivenessProbe1
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_liveness_probe_grpc import IoDragonAppsV1JobSpecTemplateSpecLivenessProbeGrpc
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_liveness_probe_tcp_socket import IoDragonAppsV1JobSpecTemplateSpecLivenessProbeTcpSocket
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_nfs import IoDragonAppsV1JobSpecTemplateSpecNfs
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_os import IoDragonAppsV1JobSpecTemplateSpecOs
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_persistent_volume_claim import IoDragonAppsV1JobSpecTemplateSpecPersistentVolumeClaim
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_photon_persistent_disk import IoDragonAppsV1JobSpecTemplateSpecPhotonPersistentDisk
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_ports import IoDragonAppsV1JobSpecTemplateSpecPorts
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_portworx_volume import IoDragonAppsV1JobSpecTemplateSpecPortworxVolume
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected import IoDragonAppsV1JobSpecTemplateSpecProjected
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected_config_map import IoDragonAppsV1JobSpecTemplateSpecProjectedConfigMap
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected_downward_api import IoDragonAppsV1JobSpecTemplateSpecProjectedDownwardAPI
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected_secret import IoDragonAppsV1JobSpecTemplateSpecProjectedSecret
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected_service_account_token import IoDragonAppsV1JobSpecTemplateSpecProjectedServiceAccountToken
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_projected_sources import IoDragonAppsV1JobSpecTemplateSpecProjectedSources
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_quobyte import IoDragonAppsV1JobSpecTemplateSpecQuobyte
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_rbd import IoDragonAppsV1JobSpecTemplateSpecRbd
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_rbd_secret_ref import IoDragonAppsV1JobSpecTemplateSpecRbdSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_readiness_gates import IoDragonAppsV1JobSpecTemplateSpecReadinessGates
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_readiness_probe import IoDragonAppsV1JobSpecTemplateSpecReadinessProbe
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_resources import IoDragonAppsV1JobSpecTemplateSpecResources
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_resources1 import IoDragonAppsV1JobSpecTemplateSpecResources1
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_scale_io import IoDragonAppsV1JobSpecTemplateSpecScaleIO
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_scale_io_secret_ref import IoDragonAppsV1JobSpecTemplateSpecScaleIOSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_secret import IoDragonAppsV1JobSpecTemplateSpecSecret
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_secret_ref import IoDragonAppsV1JobSpecTemplateSpecSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context import IoDragonAppsV1JobSpecTemplateSpecSecurityContext
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context1 import IoDragonAppsV1JobSpecTemplateSpecSecurityContext1
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context2 import IoDragonAppsV1JobSpecTemplateSpecSecurityContext2
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context2_se_linux_options import IoDragonAppsV1JobSpecTemplateSpecSecurityContext2SeLinuxOptions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context2_seccomp_profile import IoDragonAppsV1JobSpecTemplateSpecSecurityContext2SeccompProfile
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context2_sysctls import IoDragonAppsV1JobSpecTemplateSpecSecurityContext2Sysctls
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context2_windows_options import IoDragonAppsV1JobSpecTemplateSpecSecurityContext2WindowsOptions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context_capabilities import IoDragonAppsV1JobSpecTemplateSpecSecurityContextCapabilities
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context_se_linux_options import IoDragonAppsV1JobSpecTemplateSpecSecurityContextSeLinuxOptions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context_seccomp_profile import IoDragonAppsV1JobSpecTemplateSpecSecurityContextSeccompProfile
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_security_context_windows_options import IoDragonAppsV1JobSpecTemplateSpecSecurityContextWindowsOptions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_startup_probe import IoDragonAppsV1JobSpecTemplateSpecStartupProbe
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_storageos import IoDragonAppsV1JobSpecTemplateSpecStorageos
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_storageos_secret_ref import IoDragonAppsV1JobSpecTemplateSpecStorageosSecretRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_tolerations import IoDragonAppsV1JobSpecTemplateSpecTolerations
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_topology_spread_constraints import IoDragonAppsV1JobSpecTemplateSpecTopologySpreadConstraints
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_value_from import IoDragonAppsV1JobSpecTemplateSpecValueFrom
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_value_from_config_map_key_ref import IoDragonAppsV1JobSpecTemplateSpecValueFromConfigMapKeyRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_value_from_field_ref import IoDragonAppsV1JobSpecTemplateSpecValueFromFieldRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_value_from_resource_field_ref import IoDragonAppsV1JobSpecTemplateSpecValueFromResourceFieldRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_value_from_secret_key_ref import IoDragonAppsV1JobSpecTemplateSpecValueFromSecretKeyRef
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_volume_devices import IoDragonAppsV1JobSpecTemplateSpecVolumeDevices
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_volume_mounts import IoDragonAppsV1JobSpecTemplateSpecVolumeMounts
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_volumes import IoDragonAppsV1JobSpecTemplateSpecVolumes
from dragonk8s.dragon.models.io_dragon_apps_v1_job_spec_template_spec_vsphere_volume import IoDragonAppsV1JobSpecTemplateSpecVsphereVolume
from dragonk8s.dragon.models.io_dragon_apps_v1_job_status import IoDragonAppsV1JobStatus
from dragonk8s.dragon.models.io_dragon_apps_v1_job_status_conditions import IoDragonAppsV1JobStatusConditions
from dragonk8s.dragon.models.io_dragon_apps_v1_job_status_uncounted_terminated_pods import IoDragonAppsV1JobStatusUncountedTerminatedPods
from dragonk8s.dragon.models.io_dragon_apps_v1_replica_set import IoDragonAppsV1ReplicaSet
from dragonk8s.dragon.models.io_dragon_apps_v1_replica_set_list import IoDragonAppsV1ReplicaSetList
from dragonk8s.dragon.models.io_dragon_apps_v1_replica_set_spec import IoDragonAppsV1ReplicaSetSpec
from dragonk8s.dragon.models.io_dragon_apps_v1_replica_set_status import IoDragonAppsV1ReplicaSetStatus
from dragonk8s.dragon.models.io_dragon_apps_v1_replica_set_status_conditions import IoDragonAppsV1ReplicaSetStatusConditions



import logging

from dragonk8s.lib.apimachinery.pkg.labels import Selector, Set
from dragonk8s.lib.client.tools import cache
from kubernetes.client.models import V1ObjectMeta
from dragonk8s.lib.apimachinery.pkg.api import errors
from dragonk8s.lib.apimeta import apigvk


NamespaceIndex = "namespace"


def list_all(store: cache.Store, selector: Selector, append_fun):
    select_all = selector.empty()
    for m in store.list():
        if select_all:
            append_fun(m)
            continue
        metadata = m.metadata
        if selector.matches(labels=Set(data=metadata.labels)):
            append_fun(m)


def list_all_by_namespace(indexer: cache.Indexer, namespace: str, selector: Selector, append_fun):
    select_all = selector.empty()
    if namespace == "":
        for m in indexer.list():
            if select_all:
                append_fun(m)
                continue
            metadata = m.metadata
            if selector.matches(labels=Set(data=metadata.labels)):
                append_fun(m)
        return
    try:
        items = indexer.index(NamespaceIndex, V1ObjectMeta(namespace=namespace))
    except Exception as e:
        logging.warning("can not retrieve list of objects using index: %s" % str(e))
        for m in indexer.list():
            metadata = m.metadata
            if metadata.namespace == namespace and selector.matches(Set(data=metadata.labels)):
                append_fun(m)
        return
    else:
        for m in items:
            if select_all:
                append_fun(m)
                continue
            metadata = m.metadata
            if selector.matches(Set(data=metadata.labels)):
                append_fun(m)


class INamespaceLister(object):

    def list(self, selector: Selector) -> list:
        pass

    def get(self, name: str):
        pass


class ILister(object):

    def list(self, selector: Selector) -> list:
        pass

    def with_namespace(self, namespace: str) -> INamespaceLister:
        pass


class NamespaceLister(INamespaceLister):

    def __init__(self, indexer: cache.Indexer, namespace: str):
        self.indexer = indexer
        self.namespace = namespace

    def list(self, selector: Selector) -> list:
        ret = []
        list_all_by_namespace(self.indexer, self.namespace, selector, lambda o: ret.append(o))
        return ret

    def get(self, name: str):
        obj, exists = self.indexer.get_by_key("{}/{}".format(self.namespace, name))
        if not exists:
            raise errors.new_not_found(apigvk.IoDragonAppsV1ReplicaSet, name)
        return obj


class Lister(ILister):

    def __init__(self, indexer: cache.Indexer):
        self.indexer = indexer

    def list(self, selector: Selector) -> list:
        ret = []
        list_all(self.indexer, selector, lambda m: ret.append(m))
        return ret

    def with_namespace(self, namespace: str) -> INamespaceLister:
        return NamespaceLister(indexer=self.indexer, namespace=namespace)



from dragonk8s.lib.client.tools import informer, cache
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.client import CommonWatch
from dragonk8s.lib.client import lister
from kubernetes.client.models import V1Pod
from dragonk8s.lib.apimachinery.pkg import labels
from dragonk8s.lib.client.informers import factory
from dragonk8s.lib.client.informers.informer import BaseInformer


class ReplicaSetLister(lister.Lister):

    def __init__(self, indexer: cache.Indexer):
        super(ReplicaSetLister, self).__init__(indexer)

    def get_pod_replica_sets(self, pod: V1Pod) -> list:
        if len(pod.metadata.labels) == 0:
            raise Exception("no Replicasets found for pod: %s because it has no labels" % pod.metadata.name)
        l = self.with_namespace(pod.metadata.namespace).list(labels.every_thing)
        rss = []
        for rs in l:
            if rs.metadata.namespace != pod.metadata.namespace:
                continue
            selector = meta_v1.labels_selector_as_selector(rs.spec.selector)
            if selector.empty() or not selector.matches(labels.Set(pod.metadata.labels)):
                continue
            rss.append(rs)
        if len(rss) == 0:
            raise Exception("could not find ReplicaSet for pod: %s in namespace %s with labels: %s"
                            % (pod.metadata.name, pod.metadata.namespace, pod.metadata.labels))
        return rss


class ReplicaSetInformer(BaseInformer):

    def __init__(self, informer_factory: factory.SharedInformerFactory, tweak_list_options=None):
        super(ReplicaSetInformer, self).__init__(apigvk.IoDragonAppsV1ReplicaSet, informer_factory, tweak_list_options)

    def lister(self):
        return ReplicaSetLister(indexer=self.informer().get_indexer())



import time
import unittest
from kubernetes import config
from dragonk8s.lib.client.tools import record
from kubernetes.client.models.v1_event_source import V1EventSource
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from dragonk8s.lib.client.kubernetes.typed.core.v1.event_expansion import EventSinkImpl


class EventSinkImplTest(unittest.TestCase):

    def setUp(self):
        config.load_kube_config(config_file="/Users/bytedance/.kube/config")
        self.event_broadcaster = record.EventBroadcasterImpl()
        source = V1EventSource(component="mhc", host="1.2.3.4")
        self.event_recorder = self.event_broadcaster.new_recorder(source)
        self.event_broadcaster.start_recording_to_sink(EventSinkImpl())
        self.event_broadcaster.start_structured_logging()

    def test1(self):
        pod = V1Pod(
            kind="Pod",
            api_version="v1",
            metadata=V1ObjectMeta(
                name="bbb",
                namespace="kube-system"
            ),
        )
        self.event_recorder.event(pod, record.EventTypeNormal, "MHCReason", "some message....")
        time.sleep(5)

    def tearDown(self) -> None:
        print("shut down")
        self.event_broadcaster.shut_down()



from distutils.core import setup
from setuptools import find_packages

setup(
    name='dragonk8s',
    version='1.0.5',
    description="A k8s implement of Python3",
    author='ma',
    author_email='591867837@qq.com',
    url='https://dragonk8s.com',
    packages=find_packages(),
    install_requires=[
        'cachetools==5.2.0',
        'certifi==2022.9.24',
        'charset-normalizer==2.1.1',
        'google-auth==2.14.1',
        'idna==3.4',
        'kubernetes===25.3.0',
        'oauthlib==3.2.2',
        'pyasn1==0.4.8',
        'pyasn1-modules==0.2.8',
        'python-dateutil==2.8.2',
        'pytz==2022.7',
        'PyYAML==6.0',
        'requests==2.28.1',
        'requests-oauthlib==1.3.1',
        'rsa==4.9',
        'six==1.16.0',
        'urllib3==1.26.13',
        'websocket-client==1.4.2',
    ]
)